import pytest
import wpimath_test


def test_units_attributes():

    assert wpimath_test.SomeClass.s_constant == 2
    assert wpimath_test.SomeClass.ms_constant1 == 20  # the unit is ms, not seconds
    assert wpimath_test.SomeClass.ms_constant2 == 0.050
    assert wpimath_test.SomeClass.ms_constant3 == 200


def test_units_check_default_by_name1():
    sc = wpimath_test.SomeClass()

    assert sc.checkDefaultByName1(0.020) == True
    assert sc.checkDefaultByName1() == True

    with pytest.raises(RuntimeError):
        sc.checkDefaultByName1(100)


def test_units_check_default_by_name2():
    sc = wpimath_test.SomeClass()

    assert sc.checkDefaultByName2(0.050) == True
    assert sc.checkDefaultByName2() == True

    with pytest.raises(RuntimeError):
        sc.checkDefaultByName2(100)


def test_units_check_default_by_num1():
    sc = wpimath_test.SomeClass()

    assert sc.checkDefaultByNum1(0.050) == True
    assert sc.checkDefaultByNum1() == True

    with pytest.raises(RuntimeError):
        sc.checkDefaultByNum1(100)


def test_units_check_default_by_num2():
    sc = wpimath_test.SomeClass()

    assert sc.checkDefaultByNum2(0.050) == True
    assert sc.checkDefaultByNum2() == True

    with pytest.raises(RuntimeError):
        sc.checkDefaultByNum2(100)


def test_units_ft():
    assert wpimath_test.SomeClass.five_ft == 5.0


def test_units_ft2m():
    sc = wpimath_test.SomeClass()

    assert sc.ft2m(3) == 0.9144


def test_units_ms2s():
    sc = wpimath_test.SomeClass()

    assert sc.ms2s(20) == 0.020


def test_units_s2ms():
    sc = wpimath_test.SomeClass()

    assert sc.s2ms(0.2) == 200.0



# autogenerated by 'robotpy-build create-imports wpimath.trajectory.constraint wpimath.trajectory._trajectory.constraints'
from ..._controls._controls.constraint import (
    CentripetalAccelerationConstraint,
    DifferentialDriveKinematicsConstraint,
    DifferentialDriveVoltageConstraint,
    EllipticalRegionConstraint,
    MaxVelocityConstraint,
    MecanumDriveKinematicsConstraint,
    RectangularRegionConstraint,
    SwerveDrive2KinematicsConstraint,
    SwerveDrive3KinematicsConstraint,
    SwerveDrive4KinematicsConstraint,
    SwerveDrive6KinematicsConstraint,
    TrajectoryConstraint,
)

__all__ = [
    "CentripetalAccelerationConstraint",
    "DifferentialDriveKinematicsConstraint",
    "DifferentialDriveVoltageConstraint",
    "EllipticalRegionConstraint",
    "MaxVelocityConstraint",
    "MecanumDriveKinematicsConstraint",
    "RectangularRegionConstraint",
    "SwerveDrive2KinematicsConstraint",
    "SwerveDrive3KinematicsConstraint",
    "SwerveDrive4KinematicsConstraint",
    "SwerveDrive6KinematicsConstraint",
    "TrajectoryConstraint",
]



# autogenerated by 'robotpy-build create-imports wpimath.system wpimath._controls._controls.system'
from .._controls._controls.system import (
    LinearSystemId,
    LinearSystemLoop_1_1_1,
    LinearSystemLoop_2_1_1,
    LinearSystemLoop_2_2_2,
    LinearSystem_1_1_1,
    LinearSystem_1_1_2,
    LinearSystem_2_1_1,
    LinearSystem_2_1_2,
    LinearSystem_2_2_1,
    LinearSystem_2_2_2,
)

__all__ = [
    "LinearSystemId",
    "LinearSystemLoop_1_1_1",
    "LinearSystemLoop_2_1_1",
    "LinearSystemLoop_2_2_2",
    "LinearSystem_1_1_1",
    "LinearSystem_1_1_2",
    "LinearSystem_2_1_1",
    "LinearSystem_2_1_2",
    "LinearSystem_2_2_1",
    "LinearSystem_2_2_2",
]



#!/usr/bin/env python3

import io
import inspect
import pathlib
import re
import subprocess
import sys
import toml

out = {}

for f in sorted(pathlib.Path(sys.argv[1]).glob("*.h")):
    if f.name == "base.h":
        continue

    names = []

    with open(f) as fp:
        last_line = None
        for line in fp.readlines():
            line = line.strip()
            if last_line is not None:
                line = last_line + line
                last_line = None
            elif line.endswith("(") or line.endswith(","):
                last_line = line
                continue

            m = re.match(r"UNIT_ADD\(\w+, (\w+), (\w+),", line)
            if m:
                names.append((m.group(1), m.group(2)))
            else:
                m = re.match(r"UNIT_ADD_WITH_METRIC_PREFIXES\(\w+, (\w+), (\w+),", line)
                if m:
                    names.append((m.group(1), m.group(2)))
                    for i in ("nano", "micro", "milli", "kilo"):
                        names.append((f"{i}{m.group(1)}", f"{i}{m.group(2)}"))

    out_name = f"units_{f.name[:-2]}_type_caster.h"

    if names:
        if True:
            ofp = io.StringIO()

            ofp.write("#pragma once\n")
            ofp.write("\n")
            ofp.write(f"#include <units/{f.name}>\n")
            ofp.write("\n")

            ofp.write("\nnamespace pybind11 { namespace detail {\n")

            for single, double in names:

                ofp.write(
                    inspect.cleandoc(
                        f"""
                
                    template <> struct handle_type_name<units::{single}_t> {{
                    static constexpr auto name = _("{double}");
                    }};

                    template <> struct handle_type_name<units::{double}> {{
                    static constexpr auto name = _("{double}");
                    }};

                """
                    )
                )
                ofp.write("\n\n")

            ofp.write("\n}\n}\n\n")

            ofp.write(f'#include "_units_base_type_caster.h"\n')
            ofp.write("\n")

            ofp.seek(0)

            content = subprocess.check_output(
                ["clang-format"], input=ofp.getvalue(), encoding="utf-8"
            )

            with open(out_name, "w+") as fp:
                fp.write(content)
        else:
            out[out_name] = sorted(
                [f"units::{single}_t" for single, _ in names]
                + [f"units::{single}" for single, _ in names]
                + [f"units::{double}" for _, double in names]
            )


print(toml.dumps(dict(sup=out)))



from . import _init_kinematics

# autogenerated by 'robotpy-build create-imports wpimath.kinematics'
from ._kinematics import (
    ChassisSpeeds,
    DifferentialDriveKinematics,
    DifferentialDriveOdometry,
    DifferentialDriveWheelSpeeds,
    MecanumDriveKinematics,
    MecanumDriveOdometry,
    MecanumDriveWheelSpeeds,
    SwerveDrive2Kinematics,
    SwerveDrive2Odometry,
    SwerveDrive3Kinematics,
    SwerveDrive3Odometry,
    SwerveDrive4Kinematics,
    SwerveDrive4Odometry,
    SwerveDrive6Kinematics,
    SwerveDrive6Odometry,
    SwerveModuleState,
)

__all__ = [
    "ChassisSpeeds",
    "DifferentialDriveKinematics",
    "DifferentialDriveOdometry",
    "DifferentialDriveWheelSpeeds",
    "MecanumDriveKinematics",
    "MecanumDriveOdometry",
    "MecanumDriveWheelSpeeds",
    "SwerveDrive2Kinematics",
    "SwerveDrive2Odometry",
    "SwerveDrive3Kinematics",
    "SwerveDrive3Odometry",
    "SwerveDrive4Kinematics",
    "SwerveDrive4Odometry",
    "SwerveDrive6Kinematics",
    "SwerveDrive6Odometry",
    "SwerveModuleState",
]



# autogenerated by 'robotpy-build create-imports wpimath.trajectory'
from .._controls._controls.trajectory import (
    Trajectory,
    TrajectoryConfig,
    TrajectoryGenerator,
    TrajectoryParameterizer,
    TrajectoryUtil,
    TrapezoidProfile,
    TrapezoidProfileRadians,
)

__all__ = [
    "Trajectory",
    "TrajectoryConfig",
    "TrajectoryGenerator",
    "TrajectoryParameterizer",
    "TrajectoryUtil",
    "TrapezoidProfile",
    "TrapezoidProfileRadians",
]



# autogenerated by 'robotpy-build create-imports wpimath.controller wpimath._controls._controls.controller'
from .._controls._controls.controller import (
    ArmFeedforward,
    BangBangController,
    ControlAffinePlantInversionFeedforward_1_1,
    ControlAffinePlantInversionFeedforward_2_1,
    ControlAffinePlantInversionFeedforward_2_2,
    ElevatorFeedforward,
    HolonomicDriveController,
    LinearPlantInversionFeedforward_1_1,
    LinearPlantInversionFeedforward_2_1,
    LinearPlantInversionFeedforward_2_2,
    LinearQuadraticRegulator_1_1,
    LinearQuadraticRegulator_2_1,
    LinearQuadraticRegulator_2_2,
    PIDController,
    ProfiledPIDController,
    ProfiledPIDControllerRadians,
    RamseteController,
    SimpleMotorFeedforwardMeters,
)

__all__ = [
    "ArmFeedforward",
    "BangBangController",
    "ControlAffinePlantInversionFeedforward_1_1",
    "ControlAffinePlantInversionFeedforward_2_1",
    "ControlAffinePlantInversionFeedforward_2_2",
    "ElevatorFeedforward",
    "HolonomicDriveController",
    "LinearPlantInversionFeedforward_1_1",
    "LinearPlantInversionFeedforward_2_1",
    "LinearPlantInversionFeedforward_2_2",
    "LinearQuadraticRegulator_1_1",
    "LinearQuadraticRegulator_2_1",
    "LinearQuadraticRegulator_2_2",
    "PIDController",
    "ProfiledPIDController",
    "ProfiledPIDControllerRadians",
    "RamseteController",
    "SimpleMotorFeedforwardMeters",
]



import pytest

from wpimath.geometry import Pose2d, Rotation2d, Translation2d
from wpimath.kinematics import (
    ChassisSpeeds,
    SwerveDrive4Kinematics,
    SwerveDrive4Odometry,
    SwerveModuleState,
)


@pytest.fixture()
def s4():
    fl = Translation2d(12, 12)
    fr = Translation2d(12, -12)
    bl = Translation2d(-12, 12)
    br = Translation2d(-12, -12)
    return SwerveDrive4Kinematics(fl, fr, bl, br)


def test_swerve4_straightline(s4: SwerveDrive4Kinematics):
    chassisSpeeds = ChassisSpeeds(5, 0, 0)

    fl, fr, bl, br = s4.toSwerveModuleStates(chassisSpeeds)
    assert fl.speed == pytest.approx(5.0)
    assert fr.speed == pytest.approx(5.0)
    assert bl.speed == pytest.approx(5.0)
    assert br.speed == pytest.approx(5.0)

    assert fl.angle.radians() == pytest.approx(0.0)
    assert fr.angle.radians() == pytest.approx(0.0)
    assert bl.angle.radians() == pytest.approx(0.0)
    assert br.angle.radians() == pytest.approx(0.0)


def test_swerve4_normalize():
    s1 = SwerveModuleState(5)
    s2 = SwerveModuleState(6)
    s3 = SwerveModuleState(4)
    s4 = SwerveModuleState(7)

    states = SwerveDrive4Kinematics.desaturateWheelSpeeds([s1, s2, s3, s4], 5.5)

    kFactor = 5.5 / 7.0

    assert states[0].speed == pytest.approx(5.0 * kFactor)
    assert states[1].speed == pytest.approx(6.0 * kFactor)
    assert states[2].speed == pytest.approx(4.0 * kFactor)
    assert states[3].speed == pytest.approx(7.0 * kFactor)


def test_swerve4_odometry(s4: SwerveDrive4Kinematics):
    odometry = SwerveDrive4Odometry(s4, Rotation2d(0))
    odometry.resetPosition(Pose2d(), Rotation2d.fromDegrees(90))

    state = SwerveModuleState(5)

    odometry.updateWithTime(
        0,
        Rotation2d.fromDegrees(90),
        SwerveModuleState(),
        SwerveModuleState(),
        SwerveModuleState(),
        SwerveModuleState(),
    )

    pose = odometry.updateWithTime(
        0.1, Rotation2d.fromDegrees(90), state, state, state, state
    )

    assert pose.translation().x == pytest.approx(0.5)
    assert pose.translation().y == pytest.approx(0.0)
    assert pose.rotation().degrees() == pytest.approx(0)



from wpimath.trajectory import TrajectoryGenerator
from wpimath.geometry import Transform2d
import wpimath.trajectory
import wpimath.trajectory.constraint

from wpimath.geometry import Pose2d, Rotation2d, Translation2d

from wpimath.trajectory import Trajectory, TrajectoryConfig

from wpimath.trajectory.constraint import (
    MaxVelocityConstraint,
    EllipticalRegionConstraint,
    RectangularRegionConstraint,
)


def getTestTrajectory(config: TrajectoryConfig) -> Trajectory:
    # 2018 cross scale auto waypoints
    sideStart = Pose2d.fromFeet(1.54, 23.23, Rotation2d.fromDegrees(180))
    crossScale = Pose2d.fromFeet(23.7, 6.8, Rotation2d.fromDegrees(-160))

    config.setReversed(True)

    vector = [
        (
            sideStart + Transform2d(Translation2d.fromFeet(-13, 0), Rotation2d())
        ).translation(),
        (
            sideStart
            + Transform2d(
                Translation2d.fromFeet(-19.5, 5.0), Rotation2d.fromDegrees(-90)
            )
        ).translation(),
    ]

    return TrajectoryGenerator.generateTrajectory(sideStart, vector, crossScale, config)


#
# EllipticalRegionConstraint tests
#


def test_elliptical_region_constraint():
    maxVelocity = 2
    config = TrajectoryConfig.fromFps(13, 13)
    maxVelConstraint = MaxVelocityConstraint.fromFps(maxVelocity)
    regionConstraint = EllipticalRegionConstraint.fromFeet(
        Translation2d.fromFeet(5, 2.5),
        10,
        5,
        Rotation2d.fromDegrees(180),
        maxVelConstraint,
    )

    config.addConstraint(regionConstraint)

    trajectory = getTestTrajectory(config)

    exceededConstraintOutsideRegion = False
    for point in trajectory.states():
        translation = point.pose.translation()
        if translation.x_feet < 10 and translation.y_feet < 5:
            assert abs(point.velocity_fps) < maxVelocity + 0.05
        elif abs(point.velocity_fps) >= maxVelocity + 0.05:
            exceededConstraintOutsideRegion = True

    assert exceededConstraintOutsideRegion


def test_elliptical_region_is_pose_in_region():
    maxVelConstraint = MaxVelocityConstraint.fromFps(2)
    regionConstraintNoRotation = EllipticalRegionConstraint.fromFeet(
        Translation2d.fromFeet(1, 1), 2, 4, Rotation2d(0), maxVelConstraint
    )

    assert not regionConstraintNoRotation.isPoseInRegion(Pose2d.fromFeet(2.1, 1, 0))


#
# RectangularRegionConstraint tests
#


def test_rectangular_region_constraint():
    maxVelocity = 2
    config = TrajectoryConfig.fromFps(13, 13)
    maxVelConstraint = MaxVelocityConstraint.fromFps(maxVelocity)
    regionConstraint = RectangularRegionConstraint(
        Translation2d.fromFeet(1, 1),
        Translation2d.fromFeet(5, 27),
        maxVelConstraint,
    )

    config.addConstraint(regionConstraint)

    trajectory = getTestTrajectory(config)

    exceededConstraintOutsideRegion = False
    for point in trajectory.states():
        if regionConstraint.isPoseInRegion(point.pose):
            assert abs(point.velocity_fps) < maxVelocity + 0.05
        elif abs(point.velocity_fps) >= maxVelocity + 0.05:
            exceededConstraintOutsideRegion = True

    assert exceededConstraintOutsideRegion


def test_rectangular_region_is_pose_in_region():
    maxVelConstraint = MaxVelocityConstraint.fromFps(2)
    regionConstraint = RectangularRegionConstraint(
        Translation2d.fromFeet(1, 1), Translation2d.fromFeet(5, 27), maxVelConstraint
    )
    assert not regionConstraint.isPoseInRegion(Pose2d())
    assert regionConstraint.isPoseInRegion(Pose2d.fromFeet(3, 14.5, Rotation2d()))



from . import _init_geometry

# autogenerated by 'robotpy-build create-imports wpimath.geometry'
from ._geometry import Pose2d, Rotation2d, Transform2d, Translation2d, Twist2d

__all__ = ["Pose2d", "Rotation2d", "Transform2d", "Translation2d", "Twist2d"]



from . import _init_spline

# autogenerated by 'robotpy-build create-imports wpimath.spline'
from ._spline import (
    CubicHermiteSpline,
    QuinticHermiteSpline,
    Spline3,
    Spline5,
    SplineHelper,
    SplineParameterizer,
)

__all__ = [
    "CubicHermiteSpline",
    "QuinticHermiteSpline",
    "Spline3",
    "Spline5",
    "SplineHelper",
    "SplineParameterizer",
]



# autogenerated by 'robotpy-build create-imports wpimath.estimator wpimath._controls._controls.estimator'
from .._controls._controls.estimator import (
    DifferentialDrivePoseEstimator,
    ExtendedKalmanFilter_1_1_1,
    ExtendedKalmanFilter_2_1_1,
    ExtendedKalmanFilter_2_2_2,
    KalmanFilter_1_1_1,
    KalmanFilter_2_1_1,
    KalmanFilter_2_2_2,
    MecanumDrivePoseEstimator,
    SwerveDrive2PoseEstimator,
    SwerveDrive3PoseEstimator,
    SwerveDrive4PoseEstimator,
    SwerveDrive6PoseEstimator,
)

__all__ = [
    "DifferentialDrivePoseEstimator",
    "ExtendedKalmanFilter_1_1_1",
    "ExtendedKalmanFilter_2_1_1",
    "ExtendedKalmanFilter_2_2_2",
    "KalmanFilter_1_1_1",
    "KalmanFilter_2_1_1",
    "KalmanFilter_2_2_2",
    "MecanumDrivePoseEstimator",
    "SwerveDrive2PoseEstimator",
    "SwerveDrive3PoseEstimator",
    "SwerveDrive4PoseEstimator",
    "SwerveDrive6PoseEstimator",
]



from . import _init_filter

# autogenerated by 'robotpy-build create-imports wpimath.filter wpimath.filter._filter'
from ._filter import Debouncer, LinearFilter, MedianFilter, SlewRateLimiter

__all__ = ["Debouncer", "LinearFilter", "MedianFilter", "SlewRateLimiter"]



#!/usr/bin/env python3

import os
from os.path import abspath, dirname
import sys
import subprocess

if __name__ == "__main__":

    root = abspath(dirname(__file__))
    os.chdir(root)

    env = os.environ.copy()
    env["SETUPTOOLS_SCM_PRETEND_VERSION"] = "0.0.1"

    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "pip",
            "--disable-pip-version-check",
            "install",
            "--no-build-isolation",
            "-e",
            "cpp",
        ],
        env=env,
    )

    subprocess.check_call([sys.executable, "-m", "py.test"])
